import json
import unittest
import six
if six.PY3:
    from unittest.mock import Mock, patch
else:
    from mock import Mock, patch

from ..exceptions import PapermillException
from ..iorw import HttpHandler, PapermillIO


class TestPapermillIO(unittest.TestCase):
    """
    Tests for `PapermillIO`.
    """

    class FakeHandler(object):
        def __init__(self, ver):
            self.ver = ver

        def read(self, path):
            return "contents from {} for version {}".format(path, self.ver)

        def listdir(self, path):
            return ["fake", "contents"]

        def write(self, buf, path):
            return "wrote {}".format(buf)

        def pretty_path(self, path):
            return "{}/pretty/{}".format(path, self.ver)


    def setUp(self):
        self.papermill_io = PapermillIO()
        self.fake1 = self.FakeHandler(1)
        self.fake2 = self.FakeHandler(2)
        self.papermill_io.register("fake", self.fake1)

    def test_get_handler(self):
        self.assertEqual(self.papermill_io.get_handler("fake"), self.fake1)

    def test_get_local_handler(self):
        with self.assertRaises(PapermillException):
            self.papermill_io.get_handler("dne")

        self.papermill_io.register("local", self.fake2)
        self.assertEqual(self.papermill_io.get_handler("dne"), self.fake2)

    def test_register_ordering(self):
        # Should match fake1 with fake2 path
        self.assertEqual(self.papermill_io.get_handler("fake2/path"), self.fake1)

        self.papermill_io.reset()
        self.papermill_io.register("fake", self.fake1)
        self.papermill_io.register("fake2", self.fake2)

        # Should match fake1 with fake1 path, and NOT fake2 path/match
        self.assertEqual(self.papermill_io.get_handler("fake/path"), self.fake1)
        # Should match fake2 with fake2 path
        self.assertEqual(self.papermill_io.get_handler("fake2/path"), self.fake2)

    def test_read(self):
        self.assertEqual(self.papermill_io.read("fake/path"), "contents from fake/path for version 1")

    def test_listdir(self):
        self.assertEqual(self.papermill_io.listdir("fake/path"), ["fake", "contents"])

    def test_write(self):
        self.assertEqual(self.papermill_io.write("buffer", "fake/path"), "wrote buffer")

    def test_pretty_path(self):
        self.assertEqual(self.papermill_io.pretty_path("fake/path"), "fake/path/pretty/1")


class TestHttpHandler(unittest.TestCase):
    """
    Tests for `HttpHandler`.
    """

    def test_listdir(self):
        """
        Tests that listdir raises `PapermillException` saying that that
        `listdir` function is not supported.
        """
        with self.assertRaises(PapermillException) as e:
            HttpHandler.listdir('http://example.com')

        self.assertEqual(
                '{}'.format(e.exception), 'listdir is not supported by HttpHandler'
        )

    def test_read(self):
        """
        Tests that the `read` function performs a request to the giving path
        and returns the response.
        """
        path = 'http://example.com'
        text = 'request test response'

        with patch('papermill.iorw.requests.get') as mock_get:
            mock_get.return_value = Mock(text=text)
            self.assertEqual(HttpHandler.read(path), text)
            mock_get.assert_called_once_with(path)

    def test_write(self):
        """
        Tests that the `write` function performs a request a put request to
        the giving path.
        """
        path = 'http://example.com'
        buf = '{"papermill": true}'

        with patch('papermill.iorw.requests.put') as mock_put:
            HttpHandler.write(buf, path)
            mock_put.assert_called_once_with(path, json=json.loads(buf))



 # -*- coding: utf-8 -*-

from __future__ import unicode_literals, print_function

import datetime
import os
import sys

import nbformat
from jupyter_client.kernelspec import get_kernel_spec
from six import string_types, integer_types

from .conf import settings
from .exceptions import PapermillException, PapermillExecutionError
from .preprocess import PapermillExecutePreprocessor, no_tqdm, log_outputs
from .iorw import load_notebook_node, write_ipynb, read_yaml_file, get_pretty_path


def execute_notebook(notebook,
                     output,
                     parameters=None,
                     prepare_only=False,
                     kernel_name=None,
                     progress_bar=True,
                     log_output=False,
                     start_timeout=60):
    """Executes a single notebook locally.
    Args:
        notebook (str): Path to input notebook.
        output (str): Path to save executed notebook.
        parameters (dict): Arbitrary keyword arguments to pass to the notebook parameters.
        prepare_only (bool): Flag to determine if execution should occur or not.
        kernel_name (str): Name of kernel to execute the notebook against.
        progress_bar (bool): Flag for whether or not to show the progress bar.
        log_output (bool): Flag for whether or not to write notebook output to stderr.
        start_timeout (int): Duration to wait for kernel start-up.

    Returns:
         nb (NotebookNode): Executed notebook object
    """
    print("Input Notebook:  %s" % get_pretty_path(notebook), file=sys.stderr)
    print("Output Notebook: %s" % get_pretty_path(output), file=sys.stderr)
    nb = load_notebook_node(notebook)

    # Parameterize the Notebook.
    if parameters:
        _parameterize_notebook(nb, kernel_name, parameters)

    # Record specified environment variable values.
    nb.metadata.papermill[
        'environment_variables'] = _fetch_environment_variables()
    nb.metadata.papermill['output_path'] = output

    if not prepare_only:
        # Execute the Notebook.
        _execute_parameterized_notebook(nb,
                                        kernel_name,
                                        progress_bar,
                                        log_output,
                                        start_timeout)

    # Write final Notebook to disk.
    write_ipynb(nb, output)
    raise_for_execution_errors(nb, output)
    # always return notebook object
    return nb


def _parameterize_notebook(nb, kernel_name, parameters):
    """Assigned parameters into the appropiate place in the input notebook
    Args:
        nb (NotebookNode): Executable notebook object
        kernel_name (str): Name of kernel to execute the notebook against.
        parameters (dict): Arbitrary keyword arguments to pass to the notebook parameters.
    """
    # Load from a file if 'parameters' is a string.
    if isinstance(parameters, string_types):
        parameters = read_yaml_file(parameters)

    # Generate parameter content based on the kernal_name
    kernel_name = kernel_name or nb.metadata.kernelspec.name
    param_content = _build_parameter_code(kernel_name, parameters)

    param_cell_index = _find_parameters_index(nb)
    if param_cell_index >= 0:
        old_parameters = nb.cells[param_cell_index]
        old_parameters.metadata['tags'].remove('parameters')
        old_parameters.metadata['tags'].append('default parameters')

    newcell = nbformat.v4.new_code_cell(source=param_content)
    newcell.metadata['tags'] = ['parameters']

    before = nb.cells[:param_cell_index + 1]
    after = nb.cells[param_cell_index + 1:]
    nb.cells = before + [newcell] + after

    nb.metadata.papermill['parameters'] = parameters


def _execute_parameterized_notebook(nb,
                                    kernel_name=None,
                                    progress_bar=True,
                                    log_output=False,
                                    start_timeout=60):
    """Performs the actual execution of the parameterized notebook locally.
    Args:
        nb (NotebookNode): Executable notebook object.
        kernel_name (str): Name of kernel to execute the notebook against.
        progress_bar (bool): Flag for whether or not to show the progress bar.
        log_output (bool): Flag for whether or not to write notebook output to stderr.
        start_timeout (int): Duration to wait for kernel start-up.
    """
    t0 = datetime.datetime.utcnow()
    processor = PapermillExecutePreprocessor(
        timeout=None,
        startup_timeout=start_timeout,
        kernel_name=kernel_name or nb.metadata.kernelspec.name, )
    processor.progress_bar = progress_bar and not no_tqdm
    processor.log_output = log_output

    processor.preprocess(nb, {})
    t1 = datetime.datetime.utcnow()

    nb.metadata.papermill['start_time'] = t0.isoformat()
    nb.metadata.papermill['end_time'] = t1.isoformat()
    nb.metadata.papermill['duration'] = (t1 - t0).total_seconds()
    nb.metadata.papermill['exception'] = any(
        [cell.metadata.papermill.get('exception') for cell in nb.cells])


def _build_parameter_code(kernel_name, parameters):
    kernelspec = get_kernel_spec(kernel_name)
    if kernel_name in _parameter_code_builders:
        return _parameter_code_builders[kernel_name](parameters)
    elif kernelspec.language in _parameter_code_builders:
        return _parameter_code_builders[kernelspec.language](parameters)
    raise PapermillException(
        "No parameter builder functions specified for kernel '%s' or language '%s'"
        % (kernel_name, kernelspec.language))


def _find_parameters_index(nb):
    parameters_indices = []
    for idx, cell in enumerate(nb.cells):
        if "parameters" in cell.metadata.tags:
            parameters_indices.append(idx)
    if not parameters_indices:
        return -1
    elif len(parameters_indices) > 1:
        raise PapermillException("Multiple cells with parameters tag found")
    return parameters_indices[0]

def _translate_escaped_str(str_val):
    """Reusable by most interpreters"""
    if isinstance(str_val, string_types):
        str_val = str_val.encode('unicode_escape')
        if sys.version_info >= (3, 0):
            str_val = str_val.decode('utf-8')
        str_val = str_val.replace('"', r'\"')
    return '"{}"'.format(str_val)

def _translate_to_str(val):
    """Reusable by most interpreters"""
    return '{}'.format(val)

# Registry for functions that build parameter assignment code.
_parameter_code_builders = {}

# Python translaters
_translate_type_str_python = _translate_escaped_str
_translate_type_int_python = _translate_to_str
_translate_type_float_python = _translate_to_str
_translate_type_bool_python = _translate_to_str

def _translate_type_dict_python(val):
    escaped = ', '.join(["{}: {}".format(_translate_type_str_python(k), _translate_type_python(v)) for k, v in val.items()])
    return '{{{}}}'.format(escaped)

def _translate_type_list_python(val):
    escaped = ', '.join([_translate_type_python(v) for v in val])
    return '[{}]'.format(escaped)

def _translate_type_python(val):
    """Translate each of the standard json/yaml types to appropiate objects in python."""
    if isinstance(val, string_types):
        return _translate_type_str_python(val)
    # Needs to be before integer checks
    elif isinstance(val, bool):
        return _translate_type_bool_python(val)
    elif isinstance(val, integer_types):
        return _translate_type_int_python(val)
    elif isinstance(val, float):
        return _translate_type_float_python(val)
    elif isinstance(val, dict):
        return _translate_type_dict_python(val)
    elif isinstance(val, list):
        return _translate_type_list_python(val)
    # Use this generic translation as a last resort
    return _translate_escaped_str(val)

# R translaters
_translate_type_str_r = _translate_escaped_str
_translate_type_int_r = _translate_to_str
_translate_type_float_r = _translate_to_str

def _translate_type_bool_r(val):
    return 'TRUE' if val else 'FALSE'

def _translate_type_dict_r(val):
    escaped = ', '.join(["{} = {}".format(_translate_type_str_r(k), _translate_type_r(v)) for k, v in val.items()])
    return 'list({})'.format(escaped)

def _translate_type_list_r(val):
    escaped = ', '.join([_translate_type_r(v) for v in val])
    return 'list({})'.format(escaped)

def _translate_type_r(val):
    """Translate each of the standard json/yaml types to appropiate objects in R."""
    if isinstance(val, string_types):
        return _translate_type_str_r(val)
    # Needs to be before integer checks
    elif isinstance(val, bool):
        return _translate_type_bool_r(val)
    elif isinstance(val, integer_types):
        return _translate_type_int_r(val)
    elif isinstance(val, float):
        return _translate_type_float_r(val)
    elif isinstance(val, dict):
        return _translate_type_dict_r(val)
    elif isinstance(val, list):
        return _translate_type_list_r(val)
    # Use this generic translation as a last resort
    return _translate_escaped_str(val)

_translate_type_str_scala = _translate_escaped_str
_translate_type_float_scala = _translate_to_str

def _translate_type_int_scala(val):
    strval = _translate_to_str(val)
    return strval + "L" if (val > 2147483647 or val < -2147483648) else strval

def _translate_type_bool_scala(val):
    return 'true' if val else 'false'

def _translate_type_dict_scala(val):
    """Translate dicts to scala maps"""
    escaped = ', '.join(["{} -> {}".format(_translate_type_str_scala(k), _translate_type_scala(v))
                         for k, v in val.items()])
    return 'Map({})'.format(escaped)

def _translate_type_list_scala(val):
    """Translate list to scala Seq."""
    escaped = ', '.join([_translate_type_scala(v) for v in val])
    return 'Seq({})'.format(escaped)

def _translate_type_scala(val):
    """Translate each of the standard json/yaml types to appropiate objects in scala."""
    if isinstance(val, string_types):
        return _translate_type_str_scala(val)
    # Needs to be before integer checks
    elif isinstance(val, bool):
        return _translate_type_bool_scala(val)
    elif isinstance(val, integer_types):
        return _translate_type_int_scala(val)
    elif isinstance(val, float):
        return _translate_type_float_scala(val)
    elif isinstance(val, dict):
        return _translate_type_dict_scala(val)
    elif isinstance(val, list):
        return _translate_type_list_scala(val)
    # Use this generic translation as a last resort
    return _translate_escaped_str(val)

def register_param_builder(name):
    """Decorator for registering functions that write variable assignments for a given kernel or language."""

    def wrapper(func):
        _parameter_code_builders[name] = func
        return func

    return wrapper

@register_param_builder("python")
def build_python_params(parameters):
    """Writers parameter assignment code for Python kernels."""
    param_content = "# Parameters\n"
    for var, val in parameters.items():
        param_content += '{} = {}\n'.format(var, _translate_type_python(val))
    return param_content


@register_param_builder("R")
def build_r_params(parameters):
    """Writes parameters assignment code for R kernels."""
    param_content = "# Parameters\n"
    for var, val in parameters.items():
        param_content += '{} = {}\n'.format(var, _translate_type_r(val))
    return param_content

@register_param_builder("scala")
def build_scala_params(parameters):
    """Writers parameter assignment code for Python kernels."""
    param_content = "// Parameters\n"
    for var, val in parameters.items():
        param_content += 'val {} = {}\n'.format(var, _translate_type_scala(val))
    return param_content

def _fetch_environment_variables():
    ret = dict()
    for name, value in os.environ.items():
        if name in settings.ENVIRONMENT_VARIABLES:
            ret[name] = value
    return ret


ERROR_MESSAGE_TEMPLATE = (
    '<span style="color:red; font-family:Helvetica Neue, Helvetica, Arial, sans-serif; font-size:2em;">'
    "An Exception was encountered at 'In [%s]'."
    '</span>')


def raise_for_execution_errors(nb, output_path):

    error = None
    for cell in nb.cells:
        if cell.get("outputs") is None:
            continue

        for output in cell.outputs:
            if output.output_type == "error":
                error = PapermillExecutionError(
                    exec_count=cell.execution_count,
                    source=cell.source,
                    ename=output.ename,
                    evalue=output.evalue,
                    traceback=output.traceback)
                break

    if error:
        # Write notebook back out with the Error Message at the top of the Notebook.
        error_msg = ERROR_MESSAGE_TEMPLATE % str(error.exec_count)
        error_msg_cell = nbformat.v4.new_code_cell(
            source="%%html\n" + error_msg,
            outputs=[
                nbformat.v4.new_output(
                    output_type="display_data",
                    data={"text/html": error_msg})
            ],
            metadata={"inputHidden": True,
                      "hide_input": True})
        nb.cells = [error_msg_cell] + nb.cells
        write_ipynb(nb, output_path)
        raise error



from __future__ import unicode_literals, print_function

import datetime
import sys

from concurrent import futures
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors.execute import CellExecutionError

from .iorw import write_ipynb

# tqdm creates 2 globals lock which raise OSException if the execution
# environment does not have shared memory for processes, e.g. AWS Lambda
try:
    from tqdm import tqdm
    no_tqdm = False
except OSError:
    no_tqdm = True

PENDING = "pending"
RUNNING = "running"
COMPLETED = "completed"

def log_outputs(cell):
    execution_count = cell.get("execution_count")
    if not execution_count:
        return

    stderrs = []
    stdouts = []
    for output in cell.get("outputs", []):
        if output.output_type == "stream":
            if output.name == "stdout":
                stdouts.append("".join(output.text))
            elif output.name == "stderr":
                stderrs.append("".join(output.text))
        elif "data" in output and "text/plain" in output.data:
            stdouts.append(output.data['text/plain'])

    # Log stdouts
    sys.stdout.write('{:-<40}'.format("Out [%s] " % execution_count) + "\n")
    sys.stdout.write("\n".join(stdouts) + "\n")

    # Log stderrs
    sys.stderr.write('{:-<40}'.format("Out [%s] " % execution_count) + "\n")
    sys.stderr.write("\n".join(stderrs) + "\n")

class PapermillExecutePreprocessor(ExecutePreprocessor):
    """Module containing a preprocessor that executes the code cells
    and updates outputs"""

    # Copyright (c) IPython Development Team.
    # Distributed under the terms of the Modified BSD License.

    # TODO: Delete this wrapper when nbconvert allows for setting preprocessor
    # hood in a more convienent manner
    def preprocess(self, nb, resources):
        """
        Copied with one edit of super -> papermill_preprocessor from nbconvert

        Preprocess notebook executing each code cell.

        The input argument `nb` is modified in-place.

        Parameters
        ----------
        nb : NotebookNode
            Notebook being executed.
        resources : dictionary
            Additional resources used in the conversion process. For example,
            passing ``{'metadata': {'path': run_path}}`` sets the
            execution path to ``run_path``.

        Returns
        -------
        nb : NotebookNode
            The executed notebook.
        resources : dictionary
            Additional resources used in the conversion process.
        """
        path = resources.get('metadata', {}).get('path') or None

        # clear display_id map
        self._display_id_map = {}

        kernel_name = nb.metadata.get('kernelspec', {}).get('name', 'python')
        if self.kernel_name:
            kernel_name = self.kernel_name
        self.log.info("Executing notebook with kernel: %s" % kernel_name)
        self.km, self.kc = self.start_new_kernel(
            startup_timeout=self.startup_timeout,
            kernel_name=kernel_name,
            extra_arguments=self.extra_arguments,
            cwd=path)
        self.kc.allow_stdin = False
        # Parent class requires self.nb to be present temporarily during preproc
        self.nb = nb

        try:
            nb, resources = self.papermill_preprocess(nb, resources)
        finally:
            self.kc.stop_channels()
            self.km.shutdown_kernel(now=self.shutdown_kernel == 'immediate')
            # Parent class required self.nb be removed after preproc
            delattr(self, 'nb')

        return nb, resources

    def start_new_kernel(self, startup_timeout=60, kernel_name='python', **kwargs):
        km = self.kernel_manager_class(kernel_name=kernel_name)
        km.start_kernel(**kwargs)
        kc = km.client()
        kc.start_channels()
        try:
            kc.wait_for_ready(timeout=startup_timeout)
        except RuntimeError:
            kc.stop_channels()
            km.shutdown_kernel()
            raise

        return km, kc

    def papermill_preprocess(self, nb, resources):
        """
        This function acts as a replacement for the grandparent's `preprocess`
        method.

        We are doing this for the following reasons:

        1. Notebooks will stop executing when they encounter a failure but not
           raise a `CellException`. This allows us to save the notebook with the
           traceback even though a `CellExecutionError` was encountered.

        2. We want to write the notebook as cells are executed. We inject our
           logic for that here.

        3. We want to include timing and execution status information with the
           metadata of each cell.

        Parameters
        ----------
        nb : NotebookNode
            Notebook being converted
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.

        """
        output_path = nb.metadata.papermill['output_path']

        # Reset the notebook.
        for cell in nb.cells:
            # Reset the cell execution counts.
            if cell.get("execution_count") is not None:
                cell.execution_count = None

            # Clear out the papermill metadata for each cell.
            cell.metadata['papermill'] = dict(
                exception=None,
                start_time=None,
                end_time=None,
                duration=None,
                status=PENDING  # pending, running, completed
            )
            if cell.get("outputs") is not None:
                cell.outputs = []

        # Execute each cell and update the output in real time.
        with futures.ThreadPoolExecutor(max_workers=1) as executor:

            # Generate the iterator
            if self.progress_bar:
                execution_iterator = tqdm(enumerate(nb.cells), total=len(nb.cells))
            else:
                execution_iterator = enumerate(nb.cells)

            for index, cell in execution_iterator:
                cell.metadata["papermill"]["status"] = RUNNING
                future = executor.submit(write_ipynb, nb, output_path)
                t0 = datetime.datetime.utcnow()
                try:
                    if not cell.source:
                        continue

                    nb.cells[index], resources = self.preprocess_cell(
                        cell, resources, index)
                    cell.metadata['papermill']['exception'] = False
                    if self.log_output:
                        log_outputs(nb.cells[index])

                except CellExecutionError:
                    cell.metadata['papermill']['exception'] = True
                    break
                finally:
                    t1 = datetime.datetime.utcnow()
                    cell.metadata['papermill']['start_time'] = t0.isoformat()
                    cell.metadata['papermill']['end_time'] = t1.isoformat()
                    cell.metadata['papermill']['duration'] = (
                        t1 - t0).total_seconds()
                    cell.metadata['papermill']['status'] = COMPLETED
                    future.result()
        return nb, resources



import io
import os
import pytest
import shutil
import sys
import tempfile
import unittest

from functools import partial

import six
if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch
import nbformat

from nbconvert import HTMLExporter

from ..api import read_notebook
from .. import execute
from ..execute import (execute_notebook, log_outputs, _translate_type_python,
                       _translate_type_r, _translate_type_scala)
from ..exceptions import PapermillExecutionError
from . import get_notebook_path, RedirectOutput


if six.PY2:
    kernel_name = 'python2'
else:
    kernel_name = 'python3'
execute_notebook = partial(execute_notebook, kernel_name=kernel_name)


@pytest.mark.parametrize("test_input,expected", [
    ("foo", '"foo"'),
    ('{"foo": "bar"}', '"{\\"foo\\": \\"bar\\"}"'),
    ({"foo": "bar"}, '{"foo": "bar"}'),
    ({"foo": '"bar"'}, '{"foo": "\\"bar\\""}'),
    ({"foo": ["bar"]}, '{"foo": ["bar"]}'),
    ({"foo": {"bar": "baz"}}, '{"foo": {"bar": "baz"}}'),
    ({"foo": {"bar": '"baz"'}}, '{"foo": {"bar": "\\"baz\\""}}'),
    (["foo"], '["foo"]'),
    (["foo", '"bar"'], '["foo", "\\"bar\\""]'),
    ([{"foo": "bar"}], '[{"foo": "bar"}]'),
    ([{"foo": '"bar"'}], '[{"foo": "\\"bar\\""}]'),
    (12345, '12345'),
    (-54321, '-54321'),
    (1.2345, '1.2345'),
    (-5432.1, '-5432.1'),
    (True, 'True'),
    (False, 'False')
])
def test_translate_type_python(test_input, expected):
    assert _translate_type_python(test_input) == expected

@pytest.mark.parametrize("test_input,expected", [
    ("foo", '"foo"'),
    ('{"foo": "bar"}', '"{\\"foo\\": \\"bar\\"}"'),
    ({"foo": "bar"}, 'list("foo" = "bar")'),
    ({"foo": '"bar"'}, 'list("foo" = "\\"bar\\"")'),
    ({"foo": ["bar"]}, 'list("foo" = list("bar"))'),
    ({"foo": {"bar": "baz"}}, 'list("foo" = list("bar" = "baz"))'),
    ({"foo": {"bar": '"baz"'}}, 'list("foo" = list("bar" = "\\"baz\\""))'),
    (["foo"], 'list("foo")'),
    (["foo", '"bar"'], 'list("foo", "\\"bar\\"")'),
    ([{"foo": "bar"}], 'list(list("foo" = "bar"))'),
    ([{"foo": '"bar"'}], 'list(list("foo" = "\\"bar\\""))'),
    (12345, '12345'),
    (-54321, '-54321'),
    (1.2345, '1.2345'),
    (-5432.1, '-5432.1'),
    (True, 'TRUE'),
    (False, 'FALSE')
])
def test_translate_type_r(test_input, expected):
    assert _translate_type_r(test_input) == expected


@pytest.mark.parametrize("test_input,expected", [
    ("foo", '"foo"'),
    ('{"foo": "bar"}', '"{\\"foo\\": \\"bar\\"}"'),
    ({"foo": "bar"}, 'Map("foo" -> "bar")'),
    ({"foo": '"bar"'}, 'Map("foo" -> "\\"bar\\"")'),
    ({"foo": ["bar"]}, 'Map("foo" -> Seq("bar"))'),
    ({"foo": {"bar": "baz"}}, 'Map("foo" -> Map("bar" -> "baz"))'),
    ({"foo": {"bar": '"baz"'}}, 'Map("foo" -> Map("bar" -> "\\"baz\\""))'),
    (["foo"], 'Seq("foo")'),
    (["foo", '"bar"'], 'Seq("foo", "\\"bar\\"")'),
    ([{"foo": "bar"}], 'Seq(Map("foo" -> "bar"))'),
    ([{"foo": '"bar"'}], 'Seq(Map("foo" -> "\\"bar\\""))'),
    (12345, '12345'),
    (-54321, '-54321'),
    (1.2345, '1.2345'),
    (-5432.1, '-5432.1'),
    (2147483648, '2147483648L'),
    (-2147483649, '-2147483649L'),
    (True, 'true'),
    (False, 'false')
])
def test_translate_type_scala(test_input, expected):
    assert _translate_type_scala(test_input) == expected

class TestNotebookHelpers(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.notebook_name = 'simple_execute.ipynb'
        self.notebook_path = get_notebook_path(self.notebook_name)
        self.nb_test_executed_fname = os.path.join(self.test_dir, 'output_{}'.format(self.notebook_name))

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    @patch(execute.__name__ + '.PapermillExecutePreprocessor')
    def test_start_timeout(self, preproc_mock):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, start_timeout=123)
        preproc_mock.assert_called_once_with(timeout=None, startup_timeout=123, kernel_name=kernel_name)

    @patch(execute.__name__ + '.PapermillExecutePreprocessor')
    def test_default_start_timeout(self, preproc_mock):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname)
        preproc_mock.assert_called_once_with(timeout=None, startup_timeout=60, kernel_name=kernel_name)

    def test_cell_insertion(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'msg': 'Hello'})
        test_nb = read_notebook(self.nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[1].get('source').split('\n'), ['# Parameters', 'msg = "Hello"', ''])
        self.assertEqual(test_nb.parameters, {'msg': 'Hello'})

    def test_no_tags(self):
        notebook_name = 'no_parameters.ipynb'
        nb_test_executed_fname = os.path.join(self.test_dir, 'output_{}'.format(notebook_name))
        execute_notebook(get_notebook_path(notebook_name), nb_test_executed_fname, {'msg': 'Hello'})
        test_nb = read_notebook(nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[0].get('source').split('\n'), ['# Parameters', 'msg = "Hello"', ''])
        self.assertEqual(test_nb.parameters, {'msg': 'Hello'})

    def test_quoted_params(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'msg': '"Hello"'})
        test_nb = read_notebook(self.nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[1].get('source').split('\n'), ['# Parameters', r'msg = "\"Hello\""', ''])
        self.assertEqual(test_nb.parameters, {'msg': '"Hello"'})

    def test_backslash_params(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'foo': r'do\ not\ crash'})
        test_nb = read_notebook(self.nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[1].get('source').split('\n'), ['# Parameters', r'foo = "do\\ not\\ crash"', ''])
        self.assertEqual(test_nb.parameters, {'foo': r'do\ not\ crash'})

    def test_backslash_quote_params(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'foo': r'bar=\"baz\"'})
        test_nb = read_notebook(self.nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[1].get('source').split('\n'), ['# Parameters', r'foo = "bar=\\\"baz\\\""', ''])
        self.assertEqual(test_nb.parameters, {'foo': r'bar=\"baz\"'})

    def test_double_backslash_quote_params(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'foo': r'\\"bar\\"'})
        test_nb = read_notebook(self.nb_test_executed_fname)
        self.assertListEqual(test_nb.node.cells[1].get('source').split('\n'), ['# Parameters', r'foo = "\\\\\"bar\\\\\""', ''])
        self.assertEqual(test_nb.parameters, {'foo': r'\\"bar\\"'})

    def test_prepare_only(self):
        path = get_notebook_path('broken1.ipynb')
        result_path = os.path.join(self.test_dir, 'broken1.ipynb')
        # Should not raise as we don't execute the notebook at all
        execute_notebook(path, result_path, {'foo': r'do\ not\ crash'}, prepare_only=True)
        nb = read_notebook(result_path)
        self.assertEqual(nb.node.cells[0].cell_type, "code")
        self.assertEqual(nb.node.cells[0].get('source').split('\n'), ['# Parameters', r'foo = "do\\ not\\ crash"', ''])

class TestBrokenNotebook1(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test(self):
        path = get_notebook_path('broken1.ipynb')
        result_path = os.path.join(self.test_dir, 'broken1.ipynb')
        with self.assertRaises(PapermillExecutionError):
            execute_notebook(path, result_path)
        nb = read_notebook(result_path)
        self.assertEqual(nb.node.cells[0].cell_type, "code")
        self.assertEqual(nb.node.cells[0].outputs[0].output_type, "display_data")
        self.assertEqual(nb.node.cells[1].execution_count, 1)
        self.assertEqual(nb.node.cells[2].execution_count, 2)
        self.assertEqual(nb.node.cells[2].outputs[0].output_type, 'error')
        self.assertEqual(nb.node.cells[3].execution_count, None)


class TestBrokenNotebook2(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test(self):
        path = get_notebook_path('broken2.ipynb')
        result_path = os.path.join(self.test_dir, 'broken2.ipynb')
        with self.assertRaises(PapermillExecutionError):
            execute_notebook(path, result_path)
        nb = read_notebook(result_path)
        self.assertEqual(nb.node.cells[0].cell_type, "code")
        self.assertEqual(nb.node.cells[0].outputs[0].output_type, "display_data")
        self.assertEqual(nb.node.cells[1].execution_count, 1)
        self.assertEqual(nb.node.cells[2].execution_count, 2)
        self.assertEqual(nb.node.cells[2].outputs[0].output_type, 'display_data')
        self.assertEqual(nb.node.cells[2].outputs[1].output_type, 'error')
        self.assertEqual(nb.node.cells[3].execution_count, None)


class TestLogging(unittest.TestCase):

    def setUp(self):
        self.saved_stdout = sys.stdout
        self.out = io.StringIO()
        sys.stdout = self.out

    def tearDown(self):
        sys.stdout = self.saved_stdout

    def test_logging(self):

        nb = nbformat.read(get_notebook_path('test_logging.ipynb'), as_version=4)

        # stderr output
        with RedirectOutput() as redirect:
            log_outputs(nb.cells[0])
            stdout = redirect.get_stdout()
            self.assertEqual(stdout, u'Out [1] --------------------------------\n\n')
            stderr = redirect.get_stderr()
            self.assertEqual(stderr, u'Out [1] --------------------------------\nINFO:test:test text\n\n')

        # stream output
        with RedirectOutput() as redirect:
            log_outputs(nb.cells[1])
            stdout = redirect.get_stdout()
            self.assertEqual(stdout, u'Out [2] --------------------------------\nhello world\n\n')
            stderr = redirect.get_stderr()
            self.assertEqual(stderr, u'Out [2] --------------------------------\n\n')

        # text/plain output
        with RedirectOutput() as redirect:
            log_outputs(nb.cells[2])
            stdout = redirect.get_stdout()
            self.assertEqual(stdout,
                             (
                                "Out [3] --------------------------------\n"
                                "<matplotlib.axes._subplots.AxesSubplot at 0x7f8391f10290>\n"
                                "<matplotlib.figure.Figure at 0x7f830af7b350>\n"
                                )
                             )
            stderr = redirect.get_stderr()
            self.assertEqual(stderr, u'Out [3] --------------------------------\n\n')


class TestNBConvertCalls(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.notebook_name = 'simple_execute.ipynb'
        self.notebook_path = get_notebook_path(self.notebook_name)
        self.nb_test_executed_fname = os.path.join(self.test_dir, 'output_{}'.format(self.notebook_name))

        self.html_exporter = HTMLExporter()
        self.html_exporter.template_file = 'basic'

    def test_convert_output_to_html(self):
        execute_notebook(self.notebook_path, self.nb_test_executed_fname, {'msg': 'Hello'})
        test_nb = read_notebook(self.nb_test_executed_fname)

        # Ensure the notebook builds valid html without crashing
        (body, resources) = self.html_exporter.from_notebook_node(test_nb.node)
        self.assertTrue(body.startswith("\n<div"))



 # -*- coding: utf-8 -*-

from __future__ import unicode_literals

import io
import json
import os

import nbformat
import requests
import yaml

from . import __version__
from .exceptions import PapermillException
from .s3 import S3


class PapermillIO(object):
    def __init__(self):
        self.reset()

    def read(self, path):
        return self.get_handler(path).read(path)

    def write(self, buf, path):
        # Usually no return object here
        return self.get_handler(path).write(buf, path)

    def listdir(self, path):
        return self.get_handler(path).listdir(path)

    def pretty_path(self, path):
        return self.get_handler(path).pretty_path(path)

    def reset(self):
        self.__handlers = []

    def register(self, scheme, handler):
        # Keep these ordered as LIFO
        self.__handlers.insert(0, (scheme, handler))

    def get_handler(self, path):
        local_handler = None
        for scheme, handler in self.__handlers:
            if scheme == 'local':
                local_handler = handler

            if path.startswith(scheme):
                return handler

        if local_handler is None:
            raise PapermillException(
                "Could not find a registered schema handler for: {}"
                .format(path))

        return local_handler


class HttpHandler(object):
    @classmethod
    def read(cls, path):
        return requests.get(path).text

    @classmethod
    def listdir(cls, path):
        raise PapermillException('listdir is not supported by HttpHandler')

    @classmethod
    def write(cls, buf, path):
        requests.put(path, json=json.loads(buf))

    @classmethod
    def pretty_path(cls, path):
        return path


class LocalHandler(object):
    @classmethod
    def read(cls, path):
        with io.open(path, 'r') as f:
            return f.read()

    @classmethod
    def listdir(cls, path):
        return [os.path.join(path, fn) for fn in os.listdir(path)]

    @classmethod
    def write(cls, buf, path):
        with io.open(path, 'w') as f:
            f.write(buf)

    @classmethod
    def pretty_path(cls, path):
        return path


class S3Handler(object):

    keyname = None

    @classmethod
    def read(cls, path):
        s3_client = S3(keyname=cls.keyname)
        return "\n".join(s3_client.read(path))

    @classmethod
    def listdir(cls, path):
        s3_client = S3(keyname=cls.keyname)
        return s3_client.listdir(path)

    @classmethod
    def write(cls, buf, path):
        s3_client = S3(keyname=cls.keyname)
        return s3_client.cp_string(buf, path)

    @classmethod
    def pretty_path(cls, path):
        return path


# Instantiate a PapermillIO instance and register Handlers.
papermill_io = PapermillIO()
papermill_io.register("local", LocalHandler)
papermill_io.register("s3://", S3Handler)
papermill_io.register("http://", HttpHandler)
papermill_io.register("https://", HttpHandler)


def read_yaml_file(path):
    """Reads a YAML file from the location specified at 'path'."""
    return yaml.load(papermill_io.read(path))


def write_ipynb(nb, path):
    """Saves a notebook object to the specified path.
    Args:
        nb_node (nbformat.NotebookNode): Notebook object to save.
        notebook_path (str): Path to save the notebook object to.
    """
    papermill_io.write(nbformat.writes(nb), path)


def load_notebook_node(notebook_path):
    """Returns a notebook object with papermill metadata loaded from the specified path.

    Args:
        notebook_path (str): Path to the notebook file.

    Returns:
        nbformat.NotebookNode

    """
    nb = nbformat.reads(papermill_io.read(notebook_path), as_version=4)

    if not hasattr(nb.metadata, 'papermill'):
        nb.metadata['papermill'] = {
            'parameters': dict(),
            'environment_variables': dict(),
            'version': __version__
        }

    for cell in nb.cells:
        if not hasattr(cell.metadata, 'tags'):
            cell.metadata['tags'] = [
            ]  # Create tags attr if one doesn't exist.

        if not hasattr(cell.metadata, 'papermill'):
            cell.metadata['papermill'] = dict()
    return nb


def list_notebook_files(path):
    """Returns a list of all the notebook files in a directory."""
    return [p for p in papermill_io.listdir(path) if p.endswith('.ipynb')]


def get_pretty_path(path):
    return papermill_io.pretty_path(path)



from __future__ import absolute_import, division, print_function

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .api import display, record, read_notebook, read_notebooks
from .exceptions import PapermillException, PapermillExecutionError
from .execute import execute_notebook



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""""
setup.py

Note: Install ipython version based on python version.
      If python 2, install ipython 5.x (python2 is not supported in ipython 6.

"""
from __future__ import print_function
import os
import sys
from os.path import exists
from setuptools import setup

import versioneer

python_2 = sys.version_info[0] == 2
def read(fname):
    with open(fname, 'rU' if python_2 else 'r') as fhandle:
        return fhandle.read()

req_path = os.path.join(os.path.dirname('__file__'), 'requirements.txt')
required = [req.strip() for req in read(req_path).splitlines() if req.strip()]

test_req_path = os.path.join(os.path.dirname('__file__'), 'requirements-dev.txt')
test_required = [req.strip() for req in read(test_req_path).splitlines() if req.strip()]
extras_require = {
    "test": test_required
}

pip_too_old = False
pip_message = ''

try:
    import pip
    pip_version = tuple([int(x) for x in pip.__version__.split('.')[:3]])
    pip_too_old = pip_version < (9, 0, 1)
    if pip_too_old :
        # pip is too old to handle IPython deps gracefully
        pip_message = 'Your pip version is out of date. Papermill requires pip >= 9.0.1. \n'\
        'pip {} detected. Please install pip >= 9.0.1.'.format(pip.__version__)
except ImportError:
        pip_message = 'No pip detected; we were unable to import pip. \n'\
        'To use papermill, please install pip >= 9.0.1.'
except Exception:
    pass

if pip_message:
    print(pip_message, file=sys.stderr)
    sys.exit(1)


setup(name='papermill',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Parametrize and Run Jupyter Notebooks',
      author='nteract contributors',
      author_email='jupyter@googlegroups.com',
      license='BSD',
      keywords="jupyter mapreduce",
      long_description=(open('README.rst').read() if exists('README.rst') else ''),
      url='https://github.com/nteract/papermill',
      packages=['papermill'],
      install_requires=required,
      extras_require=extras_require,
      entry_points={
              'console_scripts': [
                  'papermill = papermill.cli:papermill'
              ]
          }
      )



#The following tests are purposely limited to the exposed interface by iorw.py

import os.path
import pytest
import moto
from moto import mock_s3

from ..s3 import Bucket, Prefix, Key, S3, split


@pytest.fixture
def bucket_no_service():
    """Returns a bucket instance with no services"""
    return Bucket('my_test_bucket')


@pytest.fixture
def bucket_with_service():
    """Returns a bucket instance with a service"""
    return Bucket('my_sqs_bucket', ['sqs'])


@pytest.fixture
def bucket_sqs():
    """Returns a bucket instance with a sqs service"""
    return Bucket('my_sqs_bucket', ['sqs'])

@pytest.fixture
def bucket_ec2():
    """Returns a bucket instance with a ec2 service"""
    return Bucket('my_sqs_bucket', ['ec2'])

@pytest.fixture
def bucket_multiservice():
    """Returns a bucket instance with a ec2 service"""
    return Bucket('my_sqs_bucket', ['ec2', 'sqs'])

def test_bucket_init():
    assert Bucket('my_test_bucket')
    assert Bucket('my_sqs_bucket', 'sqs')


def test_bucket_defaults():
    name = 'a bucket'

    b1 = Bucket(name)
    b2 = Bucket(name, None)

    assert b1.name == b2.name
    assert b1.service == b2.service


def test_bucket_missing_params():
    with pytest.raises(TypeError):
        Bucket(service=None)

    with pytest.raises(TypeError):
        Bucket()


def test_bucket_list(bucket_sqs):
    # prefix_test = ''
    # assert bucket_sqs.list(prefix_test)
    #
    # prefix_test = 'abc'
    # assert bucket_sqs.list(prefix_test) is None
    #
    # prefix_test = 'ec2'
    # assert bucket_sqs.list(prefix_test) is None
    #
    # prefix_test = 'sqs'
    # assert bucket_sqs.list(prefix_test)
    pass

def test_prefix_init():
    pass


def test_prefix_defaults():
    bucket = Bucket('my data pool')
    name = 'bigdata bucket'

    p1 = Prefix(bucket, name)
    p2 = Prefix(bucket, name, None)
    assert p1.name == p2.name
    assert p1.service == p2.service


def test_prefix_str(bucket_sqs):
    p1 = Prefix(bucket_sqs, 'sqs_prefix_test', 'sqs')
    assert str(p1) == 's3://' + str(bucket_sqs) + '/sqs_prefix_test'


def test_prefix_repr(bucket_sqs):
    p1 = Prefix(bucket_sqs, 'sqs_prefix_test', 'sqs')
    assert p1.__repr__


def test_key_init():
    pass


def test_key_defaults():
    bucket = Bucket('my data pool')
    name = 'bigdata bucket'

    k1 = Key(bucket, name)
    k2 = Key(bucket, name, None, None, None, None, None)
    assert k1.size == k2.size
    assert k1.etag == k2.etag
    assert k1.storage_class == k2.storage_class
    assert k1.service == k2.service
    assert k1.is_prefix is False


@mock_s3
def test_s3_defaults():
    s1 = S3()
    s2 = S3(None, None, None, 'us-east-1')
    assert s1.session == s2.session
    assert s1.client == s2.client
    assert s1.s3 == s2.s3


@pytest.mark.parametrize("value,expected", [
    ('s3://foo/bar/baz', ['foo', 'bar/baz']),
    ('s3://foo/bar/baz/', ['foo', 'bar/baz/']),
    ('s3://foo', ['foo', '']),
    ('s3://', ['', '']),
    ('s3:///', ['', '']),
])
def test_split_success(value, expected):
    assert (split(value)) == expected


def test_split_error():

    with pytest.raises(ValueError):
        split('foo/bar/baz')

    with pytest.raises(ValueError):
        split('https://foo/bar/baz')


local_dir = os.path.dirname(os.path.abspath(__file__))
test_bucket_name = 'test-pm-bucket'
test_string = 'Hello'
test_file_path = 'notebooks/s3/s3_in/s3-simple_notebook.ipynb'

with open(os.path.join(local_dir, test_file_path)) as f:
    test_nb_content = f.read()

no_empty_lines = lambda s: "\n".join([l for l in s.split('\n') if len(l) > 0])
test_clean_nb_content = no_empty_lines(test_nb_content)

read_from_gen = lambda g: "\n".join(g)


@pytest.yield_fixture
def s3_client():
    mock_s3 = moto.mock_s3()
    mock_s3.start()
    import boto3
    client = boto3.client('s3')
    client.create_bucket(Bucket=test_bucket_name)
    client.put_object(Bucket=test_bucket_name,
                      Key=test_file_path,
                      Body=test_nb_content)
    yield S3()
    try:
        client.delete_object(Bucket=test_bucket_name,
                             Key=test_file_path)
        client.delete_object(Bucket=test_bucket_name,
                             Key=test_file_path+'.txt')
    except:
        pass
    mock_s3.stop()


def test_s3_read(s3_client):
    s3_path = "s3://{}/{}".format(test_bucket_name, test_file_path)
    data = read_from_gen(s3_client.read(s3_path))
    assert data == test_clean_nb_content

def test_s3_write(s3_client):
    s3_path = "s3://{}/{}.txt".format(test_bucket_name, test_file_path)
    s3_client.cp_string(test_string, s3_path)

    data = read_from_gen(s3_client.read(s3_path))
    assert data == test_string


def test_s3_overwrite(s3_client):
    s3_path = "s3://{}/{}".format(test_bucket_name, test_file_path)
    s3_client.cp_string(test_string, s3_path)

    data = read_from_gen(s3_client.read(s3_path))
    assert data == test_string


def test_s3_listdir(s3_client):
    dir_name = os.path.dirname(test_file_path)
    s3_dir = "s3://{}/{}".format(test_bucket_name, dir_name)
    s3_path = "s3://{}/{}".format(test_bucket_name, test_file_path)
    dir_listings = s3_client.listdir(s3_dir)
    assert len(dir_listings) == 1
    assert s3_path in dir_listings



 # -*- coding: utf-8 -*-
 
"""Manages the Settings for Papermill"""
from __future__ import unicode_literals


class Settings(object):

    __default_settings = {'ENVIRONMENT_VARIABLES': []}

    def __init__(self):

        for name, value in self.__default_settings.items():
            setattr(self, name, value)


settings = Settings()
